"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const metadataFiles_1 = __importDefault(require("../../metadata/metadataFiles"));
const simplegit = require("simple-git/promise");
const xml2js = __importStar(require("xml2js"));
const path = __importStar(require("path"));
const fs = __importStar(require("fs"));
const rimraf = require("rimraf");
const metadataInfo_1 = require("../../metadata/metadataInfo");
const fileutils_1 = __importDefault(require("../../../utils/fileutils"));
const _ = __importStar(require("lodash"));
const profileDiff_1 = __importDefault(require("./profileDiff"));
const permsetDiff_1 = __importDefault(require("./permsetDiff"));
const workflowDiff_1 = __importDefault(require("./workflowDiff"));
const sharingRuleDiff_1 = __importDefault(require("./sharingRuleDiff"));
const customLabelsDiff_1 = __importDefault(require("./customLabelsDiff"));
const diffUtil_1 = __importDefault(require("./diffUtil"));
const command_1 = require("@salesforce/command");
const sfpowerkit_1 = require("../../../sfpowerkit");
command_1.core.Messages.importMessagesDirectory(__dirname);
const messages = command_1.core.Messages.loadMessages("sfpowerkit", "project_diff");
const deleteNotSupported = ["RecordType"];
const git = simplegit();
const unsplitedMetadataExtensions = metadataInfo_1.UNSPLITED_METADATA.map(elem => {
    return elem.sourceExtension;
});
const permissionExtensions = metadataInfo_1.PROFILE_PERMISSIONSET_EXTENSION.map(elem => {
    return elem.sourceExtension;
});
class DiffImpl {
    constructor(revisionFrom, revisionTo, isDestructive) {
        this.revisionFrom = revisionFrom;
        this.revisionTo = revisionTo;
        this.isDestructive = isDestructive;
        if (this.revisionTo == null || this.revisionTo.trim() === "") {
            this.revisionTo = "HEAD";
        }
        if (this.revisionFrom == null) {
            this.revisionFrom = "";
        }
        this.destructivePackageObjPost = new Array();
        this.destructivePackageObjPre = new Array();
        this.resultOutput = [];
    }
    async build(diffFilePath, encoding, outputFolder) {
        rimraf.sync(outputFolder);
        //const sepRegex=/\t| |\n/;
        const sepRegex = /\n|\r/;
        let data = "";
        if (diffFilePath !== null && diffFilePath !== "") {
            data = fs.readFileSync(diffFilePath, encoding);
        }
        else {
            //check if same commit
            const commitFrom = await git.raw([
                "rev-list",
                "-n",
                "1",
                this.revisionFrom
            ]);
            const commitTo = await git.raw(["rev-list", "-n", "1", this.revisionTo]);
            const headCommit = await git.raw(["rev-list", "-n", "1", "HEAD"]);
            if (commitFrom === commitTo) {
                throw new Error(messages.getMessage("sameCommitErrorMessage"));
            }
            data = await git.diff(["--raw", this.revisionFrom, this.revisionTo]);
            sfpowerkit_1.SFPowerkit.log(`Input Param: From: ${this.revisionFrom}  To: ${this.revisionTo} `, sfpowerkit_1.LoggerLevel.INFO);
            sfpowerkit_1.SFPowerkit.log(`SHA Found From: ${commitFrom} To:  ${commitTo} `, sfpowerkit_1.LoggerLevel.INFO);
            sfpowerkit_1.SFPowerkit.log(data, sfpowerkit_1.LoggerLevel.TRACE);
        }
        let content = data.split(sepRegex);
        let diffFile = await diffUtil_1.default.parseContent(content);
        await diffUtil_1.default.fetchFileListRevisionTo(this.revisionTo);
        let filesToCopy = diffFile.addedEdited;
        let deletedFiles = diffFile.deleted;
        deletedFiles = deletedFiles.filter(deleted => {
            let found = false;
            let deletedMetadata = metadataFiles_1.default.getFullApiNameWithExtension(deleted.path);
            for (let i = 0; i < filesToCopy.length; i++) {
                let addedOrEdited = metadataFiles_1.default.getFullApiNameWithExtension(filesToCopy[i].path);
                if (deletedMetadata === addedOrEdited) {
                    found = true;
                    break;
                }
            }
            return !found;
        });
        if (fs.existsSync(outputFolder) == false) {
            fs.mkdirSync(outputFolder);
        }
        sfpowerkit_1.SFPowerkit.log("Files to be copied", sfpowerkit_1.LoggerLevel.DEBUG);
        sfpowerkit_1.SFPowerkit.log(filesToCopy, sfpowerkit_1.LoggerLevel.DEBUG);
        if (filesToCopy && filesToCopy.length > 0) {
            for (var i = 0; i < filesToCopy.length; i++) {
                let filePath = filesToCopy[i].path;
                let matcher = filePath.match(metadataInfo_1.SOURCE_EXTENSION_REGEX);
                let extension = "";
                if (matcher) {
                    extension = matcher[0];
                }
                else {
                    extension = path.parse(filePath).ext;
                }
                if (unsplitedMetadataExtensions.includes(extension)) {
                    //handle unsplited files
                    await this.handleUnsplittedMetadata(filesToCopy[i], outputFolder);
                }
                else {
                    await diffUtil_1.default.copyFile(filePath, outputFolder);
                    sfpowerkit_1.SFPowerkit.log(`Copied file ${filePath} to ${outputFolder}`, sfpowerkit_1.LoggerLevel.DEBUG);
                }
            }
        }
        if (this.isDestructive) {
            sfpowerkit_1.SFPowerkit.log("Creating Destructive Manifest..", sfpowerkit_1.LoggerLevel.INFO);
            await this.createDestructiveChanges(deletedFiles, outputFolder);
        }
        sfpowerkit_1.SFPowerkit.log(`Generating output summary`, sfpowerkit_1.LoggerLevel.INFO);
        this.buildOutput(outputFolder);
        if (this.resultOutput.length > 0) {
            try {
                diffUtil_1.default.copyFile(".forceignore", outputFolder);
            }
            catch (e) {
                if (e.code !== "EPERM") {
                    throw e;
                }
            }
            try {
                diffUtil_1.default.copyFile("sfdx-project.json", outputFolder);
            }
            catch (e) {
                if (e.code !== "EPERM") {
                    throw e;
                }
            }
        }
        return this.resultOutput;
    }
    buildOutput(outputFolder) {
        let metadataFiles = new metadataFiles_1.default();
        metadataFiles.loadComponents(outputFolder, false);
        let keys = Object.keys(metadataInfo_1.METADATA_INFO);
        let excludedFiles = _.difference(unsplitedMetadataExtensions, permissionExtensions);
        keys.forEach(key => {
            if (metadataInfo_1.METADATA_INFO[key].files && metadataInfo_1.METADATA_INFO[key].files.length > 0) {
                metadataInfo_1.METADATA_INFO[key].files.forEach(filePath => {
                    let matcher = filePath.match(metadataInfo_1.SOURCE_EXTENSION_REGEX);
                    let extension = "";
                    if (matcher) {
                        extension = matcher[0];
                    }
                    else {
                        extension = path.parse(filePath).ext;
                    }
                    if (!excludedFiles.includes(extension)) {
                        let name = fileutils_1.default.getFileNameWithoutExtension(filePath, metadataInfo_1.METADATA_INFO[key].sourceExtension);
                        if (metadataInfo_1.METADATA_INFO[key].isChildComponent) {
                            let fileParts = filePath.split(path.sep);
                            let parentName = fileParts[fileParts.length - 3];
                            name = parentName + "." + name;
                        }
                        this.resultOutput.push({
                            action: "Deploy",
                            metadataType: metadataInfo_1.METADATA_INFO[key].xmlName,
                            componentName: name,
                            path: filePath
                        });
                    }
                });
            }
        });
        return this.resultOutput;
    }
    async handleUnsplittedMetadata(diffFile, outputFolder) {
        let content1 = "";
        let content2 = "";
        try {
            if (diffFile.revisionFrom !== "0000000") {
                content1 = await git.show(["--raw", diffFile.revisionFrom]);
            }
        }
        catch (e) { }
        try {
            if (diffFile.revisionTo !== "0000000") {
                content2 = await git.show(["--raw", diffFile.revisionTo]);
            }
        }
        catch (e) { }
        fileutils_1.default.mkDirByPathSync(path.join(outputFolder, path.parse(diffFile.path).dir));
        if (diffFile.path.endsWith(metadataInfo_1.METADATA_INFO.Workflow.sourceExtension)) {
            //Workflow
            let baseName = path.parse(diffFile.path).base;
            let objectName = baseName.split(".")[0];
            await workflowDiff_1.default.generateWorkflowXml(content1, content2, path.join(outputFolder, diffFile.path), objectName, this.destructivePackageObjPost, this.resultOutput, this.isDestructive);
        }
        if (diffFile.path.endsWith(metadataInfo_1.METADATA_INFO.SharingRules.sourceExtension)) {
            let baseName = path.parse(diffFile.path).base;
            let objectName = baseName.split(".")[0];
            await sharingRuleDiff_1.default.generateSharingRulesXml(content1, content2, path.join(outputFolder, diffFile.path), objectName, this.destructivePackageObjPost, this.resultOutput, this.isDestructive);
        }
        if (diffFile.path.endsWith(metadataInfo_1.METADATA_INFO.CustomLabels.sourceExtension)) {
            await customLabelsDiff_1.default.generateCustomLabelsXml(content1, content2, path.join(outputFolder, diffFile.path), this.destructivePackageObjPost, this.resultOutput, this.isDestructive);
        }
        if (diffFile.path.endsWith(metadataInfo_1.METADATA_INFO.Profile.sourceExtension)) {
            //Deploy only what changed
            if (content1 === "") {
                await diffUtil_1.default.copyFile(diffFile.path, outputFolder);
                sfpowerkit_1.SFPowerkit.log(`Copied file ${diffFile.path} to ${outputFolder}`, sfpowerkit_1.LoggerLevel.DEBUG);
            }
            else if (content2 === "") {
                //The profile is deleted or marked as renamed.
                //Delete the renamed one
                let profileType = _.find(this.destructivePackageObjPost, function (metaType) {
                    return metaType.name === metadataInfo_1.METADATA_INFO.Profile.xmlName;
                });
                if (profileType === undefined) {
                    profileType = {
                        name: metadataInfo_1.METADATA_INFO.Profile.xmlName,
                        members: []
                    };
                    this.destructivePackageObjPost.push(profileType);
                }
                let baseName = path.parse(diffFile.path).base;
                let profileName = baseName.split(".")[0];
                profileType.members.push(profileName);
            }
            else {
                await profileDiff_1.default.generateProfileXml(content1, content2, path.join(outputFolder, diffFile.path));
            }
        }
        if (diffFile.path.endsWith(metadataInfo_1.METADATA_INFO.PermissionSet.sourceExtension)) {
            let sourceApiVersion = await sfpowerkit_1.SFPowerkit.getApiVersion();
            if (content1 === "") {
                await diffUtil_1.default.copyFile(diffFile.path, outputFolder);
                sfpowerkit_1.SFPowerkit.log(`Copied file ${diffFile.path} to ${outputFolder}`, sfpowerkit_1.LoggerLevel.DEBUG);
            }
            else if (sourceApiVersion <= 39.0) {
                // in API 39 and erliar PermissionSet deployment are merged. deploy only what changed
                if (content2 === "") {
                    //Deleted permissionSet
                    let permsetType = _.find(this.destructivePackageObjPost, function (metaType) {
                        return metaType.name === metadataInfo_1.METADATA_INFO.PermissionSet.xmlName;
                    });
                    if (permsetType === undefined) {
                        permsetType = {
                            name: metadataInfo_1.METADATA_INFO.PermissionSet.xmlName,
                            members: []
                        };
                        this.destructivePackageObjPost.push(permsetType);
                    }
                    let baseName = path.parse(diffFile.path).base;
                    let permsetName = baseName.split(".")[0];
                    permsetType.members.push(permsetName);
                }
                else {
                    await permsetDiff_1.default.generatePermissionsetXml(content1, content2, path.join(outputFolder, diffFile.path));
                }
            }
            else {
                //PermissionSet deployment override in the target org
                //So deploy the whole file
                await diffUtil_1.default.copyFile(diffFile.path, outputFolder);
                sfpowerkit_1.SFPowerkit.log(`Copied file ${diffFile.path} to ${outputFolder}`, sfpowerkit_1.LoggerLevel.DEBUG);
            }
        }
    }
    async createDestructiveChanges(filePaths, outputFolder) {
        if (_.isNil(this.destructivePackageObjPost)) {
            this.destructivePackageObjPost = new Array();
        }
        else {
            this.destructivePackageObjPost = this.destructivePackageObjPost.filter(metaType => {
                return !_.isNil(metaType.members) && metaType.members.length > 0;
            });
        }
        this.destructivePackageObjPre = new Array();
        //returns root, dir, base and name
        for (let i = 0; i < filePaths.length; i++) {
            let filePath = filePaths[i].path;
            let matcher = filePath.match(metadataInfo_1.SOURCE_EXTENSION_REGEX);
            let extension = "";
            if (matcher) {
                extension = matcher[0];
            }
            else {
                extension = path.parse(filePath).ext;
            }
            if (unsplitedMetadataExtensions.includes(extension)) {
                //handle unsplited files
                await this.handleUnsplittedMetadata(filePaths[i], outputFolder);
                continue;
            }
            let parsedPath = path.parse(filePath);
            let filename = parsedPath.base;
            let name = metadataInfo_1.MetadataInfo.getMetadataName(filePath);
            if (name) {
                if (!metadataFiles_1.default.isCustomMetadata(filePath, name)) {
                    // avoid to generate destructive for Standard Components
                    //Support on Custom Fields and Custom Objects for now
                    this.resultOutput.push({
                        action: "Skip ",
                        componentName: metadataFiles_1.default.getMemberNameFromFilepath(filePath, name),
                        metadataType: "StandardField/CustomMetadata",
                        path: "--"
                    });
                    continue;
                }
                let member = metadataFiles_1.default.getMemberNameFromFilepath(filePath, name);
                if (name === metadataInfo_1.METADATA_INFO.CustomField.xmlName) {
                    let isFormular = await diffUtil_1.default.isFormulaField(filePaths[i]);
                    if (isFormular) {
                        this.destructivePackageObjPre = this.buildDestructiveTypeObj(this.destructivePackageObjPre, name, member);
                        sfpowerkit_1.SFPowerkit.log(`${filePath} ${metadataFiles_1.default.isCustomMetadata(filePath, name)}`, sfpowerkit_1.LoggerLevel.DEBUG);
                        this.resultOutput.push({
                            action: "Delete",
                            componentName: member,
                            metadataType: name,
                            path: "Manual Intervention Required"
                        });
                    }
                    else {
                        this.destructivePackageObjPost = this.buildDestructiveTypeObj(this.destructivePackageObjPost, name, member);
                    }
                    sfpowerkit_1.SFPowerkit.log(`${filePath} ${metadataFiles_1.default.isCustomMetadata(filePath, name)}`, sfpowerkit_1.LoggerLevel.DEBUG);
                    this.resultOutput.push({
                        action: "Delete",
                        componentName: member,
                        metadataType: name,
                        path: "destructiveChanges.xml"
                    });
                }
                else {
                    if (!deleteNotSupported.includes(name)) {
                        this.destructivePackageObjPost = this.buildDestructiveTypeObj(this.destructivePackageObjPost, name, member);
                        this.resultOutput.push({
                            action: "Delete",
                            componentName: member,
                            metadataType: name,
                            path: "destructiveChanges.xml"
                        });
                    }
                    else {
                        //add the component in the manual action list
                        // TODO
                    }
                }
            }
        }
        // this.writeDestructivechanges(
        //   this.destructivePackageObjPre,
        //   outputFolder,
        //   "destructiveChangesPre.xml"
        // );
        this.writeDestructivechanges(this.destructivePackageObjPost, outputFolder, "destructiveChanges.xml");
    }
    writeDestructivechanges(destrucObj, outputFolder, fileName) {
        //ensure unique component per type
        for (let i = 0; i < destrucObj.length; i++) {
            destrucObj[i].members = _.uniq(destrucObj[i].members);
        }
        destrucObj = destrucObj.filter(metaType => {
            return metaType.members && metaType.members.length > 0;
        });
        if (destrucObj.length > 0) {
            let dest = {
                Package: {
                    $: {
                        xmlns: "http://soap.sforce.com/2006/04/metadata"
                    },
                    types: destrucObj
                }
            };
            let destructivePackageName = fileName;
            let filepath = path.join(outputFolder, destructivePackageName);
            let builder = new xml2js.Builder();
            let xml = builder.buildObject(dest);
            fs.writeFileSync(filepath, xml);
        }
    }
    buildDestructiveTypeObj(destructiveObj, name, member) {
        let typeIsPresent = false;
        for (let i = 0; i < destructiveObj.length; i++) {
            if (destructiveObj[i].name === name) {
                typeIsPresent = true;
                destructiveObj[i].members.push(member);
                break;
            }
        }
        let typeNode;
        if (typeIsPresent === false) {
            typeNode = {
                name: name,
                members: [member]
            };
            destructiveObj.push(typeNode);
        }
        return destructiveObj;
    }
}
exports.default = DiffImpl;
//# sourceMappingURL=diffImpl.js.map